#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 18:34:18 2021

@author: sandro
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import (Dirichlet, Categorical)
import math
import numpy as np
from time import time
from matplotlib import pyplot as plt
from MVN_Models import MixtureOfMVNs, MultivariateNormalDiag

def illustrate_variance(mu, variance):
    
    variance = variance.numpy()
    variance_circles = []
    for i in range(len(mu)):
        circ_x = []
        circ_y = []
        for theta in range(361):
            x = 1.5*variance[i][0]*math.cos(theta/360*2*math.pi) + mu[i][0]
            y = 1.5*variance[i][1]*math.sin(theta/360*2*math.pi) + mu[i][1]
            circ_x.append(x)
            circ_y.append(y)
        variance_circles.append([circ_x, circ_y])
    
    return variance_circles
    
    
def plot_centroids(dataset, mu_true=None, mu_model=None, mu_bench=None, 
                   var_true=None, var_model=None, var_bench=None):
    
    # centroids = model.get_centroids()
    # centroids = centroids.transpose()
    global_plot_limit = 4
    # now plot the centroids
    fig, ax = plt.subplots(figsize=(10,10))
    ax.set_title('MVN Sample', fontsize=25)
    ax.set_xlim([-global_plot_limit,global_plot_limit])
    ax.set_ylim([-global_plot_limit,global_plot_limit])
    # ax.axes.xaxis.set_visible(False)
    # ax.axes.yaxis.set_visible(False)
        
    # If a dataset is given, then also plot dataset points
    if dataset != None:
        points = dataset.numpy()
        points = points.transpose()
        amount_of_points = points.shape[1]
        alpha = 1200.0/amount_of_points
        print(alpha)
        alpha = min(1, alpha)
        plt.scatter(points[0], points[1], c='k', alpha=alpha, linewidths=0)
        
    if mu_true != None:
        mu = mu_true.numpy()
        mu = mu.transpose()
        plt.scatter(mu[0], mu[1], c='b', s=200, linewidths=5, marker='x')
        if var_true != None:
            variance_circles = illustrate_variance(mu_true, var_true)
            for circle in variance_circles:
                plt.plot(circle[0], circle[1], c='b', linewidth=3)
        
    if mu_model != None:
        mu = mu_model.numpy()
        mu = mu.transpose()
        plt.scatter(mu[0], mu[1], c='r', linewidths=0, marker='x')
        if var_model != None:
            variance_circles = illustrate_variance(mu_model, var_model)
            for circle in variance_circles:
                plt.plot(circle[0], circle[1], c='b', linewidth=3)
        
    if mu_bench != None:
        mu = mu_bench.numpy()
        mu = mu.transpose()
        plt.scatter(mu[0], mu[1], c='g', linewidths=0, marker='x')
        if var_bench != None:
            variance_circles = illustrate_variance(mu_bench, var_bench)
            for circle in variance_circles:
                plt.plot(circle[0], circle[1], c='b', linewidth=3)
        
    # Move left y-axis and bottim x-axis to centre, passing through (0,0)
    ax.spines['left'].set_position('center')
    ax.spines['bottom'].set_position('center')
    
    # Eliminate upper and right axes
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    
    # Show ticks in the left and lower axes only
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    
        # plt.scatter(centroids[0], centroids[1], c='g', marker="x", s=350, linewidths=6)
    
    # if a folder where to save the figure is specified, thenn the figure will be saved
    # if saving_directory != None:
    #     fig_name = 'Centroids.png'
    #     directory = os.path.join(figure_main_directory, saving_directory)
    #     os.makedirs(directory, exist_ok=True)
    #     path_to_save = os.path.join(directory, fig_name)
    #     plt.savefig(path_to_save)
    plt.savefig('points_vis.png')
    
    plt.show()
    
D = 2
B = 1
N = 10000
K = 4
mvn = MultivariateNormalDiag(D, sigma=0.6)
mog = MixtureOfMVNs(mvn)
    
X, labels, pi, (mu, sigma) = mog.sample(B, N, K, return_gt=True)

print(mu.squeeze())
print(X.squeeze())
plot_centroids(X.squeeze(), mu_true=mu.squeeze(), var_true=sigma.squeeze())