#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  7 14:54:35 2022

@author: sandro
"""

import numpy as np

def pbc_diff(pos1, pos2, box_size=6):
    diff = pos1 - pos2
    # Periodic boundry conditions
    diff[diff > box_size/2] = diff[diff > box_size/2] - box_size 
    diff[diff <= -box_size/2] = diff[diff <= -box_size/2] + box_size
    return diff

def kinetic_energy(states):
    momentum = np.multiply(states[:, -2:], states[:, 0][:, np.newaxis])

    kinetic_energy = np.sum(np.divide(np.square(momentum), (2*states[:, 0][:, np.newaxis])))

    return kinetic_energy

def potential_energy(states, physical_const=2, box_size=6, softening=False,
                     softening_radius=0.001):
    n_particles = states.shape[0]

    has_charge = (states.shape[1] == 6) # check if data is from Coulomb simulation

    if n_particles > 1000:
        # If we have many particles iteration is faster
        potential_energy = np.zeros(1)

        for i in range(n_particles):
            diff = pbc_diff(states[:, -4:-2], states[i, -4:-2], box_size=box_size)
            diff[i] = np.zeros((2)) # make sure current particle is not attracted to itself
            distance = np.linalg.norm(diff, axis=1)
            distance[i] = 1 # avoid division by 0 error
            if has_charge:
                m2 =  - np.multiply(states[:, 1], states[i, 1]) # q * q[i]
            else:
                m2 = np.multiply(states[:, 0], states[i, 0]) # m * m[i]
            m2[i] = np.zeros((1)) # make sure current particle is not attracted to itself
            if softening:
                potential_energy -= np.sum((np.divide((physical_const * m2),
                                            np.sqrt(distance**2 + softening_radius**2))))
            else:
                potential_energy -= np.sum((np.divide((physical_const * m2), distance)))
        self_ids = np.arange(n_particles)

        return potential_energy/2
    else:
        self_ids = np.arange(n_particles)

        diff = pbc_diff(np.repeat(states[np.newaxis, :, -4:-2], n_particles, axis=0),
                        states[:, np.newaxis, -4:-2], box_size=box_size)
        distance = np.linalg.norm(diff, axis=2)
        distance[self_ids, self_ids] = 1 # avoid division by 0 error
        if has_charge:
            m2 = - np.multiply(np.repeat(states[np.newaxis, :, 1], n_particles, axis=0),
                               states[:, np.newaxis, 1]) # q * q[i]
        else:
            m2 = np.multiply(np.repeat(states[np.newaxis, :, 0], n_particles, axis=0),
                             states[:, np.newaxis, 0]) # m * m[i]
        m2[self_ids, self_ids] = np.zeros((1)) # make sure particle is not attracted to itself
        if softening:
            potential_energy = -np.sum((np.divide((physical_const * m2),
                                            np.sqrt(distance**2 + softening_radius**2))))
        else:
            potential_energy = -np.sum((np.divide((physical_const * m2), distance)))

    return potential_energy/2 # we double count

def hamiltonian(states, physical_const=2, box_size=6, softening=False, softening_radius=0.001):
    # Total energy of the system
    return kinetic_energy(states) + potential_energy(states, physical_const, 
                box_size=box_size, softening=softening, softening_radius=softening_radius)