#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 20:10:37 2021

@author: sandro
"""

import torch
import torch.nn as nn

from Tree_attention_scaling_simple_tree_no_sparse.tree_attention_models import\
    Tree_Set_Transformer_Encoder, Set_Transformer_Encoder, Full_Set_Transformer_Encoder,\
        Enhanced_Tree_Set_Transformer_Encoder, Standard_Tree_Transformer,\
            Deep_Standard_Tree_Transformer, Shallow_Standard_Tree_Transformer,\
                Standard_Transformer, Set_Transformer_Encoder_deep

device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

class Standard_Transformer_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Standard_Transformer_Segmentation_Network, self).__init__()
        self.enc = Standard_Tree_Transformer(dim_input, num_classes, num_leafs,
                        ln=ln, max_tree_construction=max_tree_construction, tree=tree,
                        num_trees=num_trees, dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class Deep_Standard_Transformer_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Deep_Standard_Transformer_Segmentation_Network, self).__init__()
        self.enc = Deep_Standard_Tree_Transformer(dim_input, num_classes, num_leafs,
                        ln=ln, max_tree_construction=max_tree_construction, tree=tree,
                        num_trees=num_trees, dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class Shallow_Standard_Transformer_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Shallow_Standard_Transformer_Segmentation_Network, self).__init__()
        self.enc = Shallow_Standard_Tree_Transformer(dim_input, num_classes, num_leafs,
                        ln=ln, max_tree_construction=max_tree_construction, tree=tree,
                        num_trees=num_trees, dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class Standard_Transformer_notree(nn.Module):
    
    def __init__(self, dim_input, num_classes, dim_hidden, ln=False):
        super(Standard_Transformer_notree, self).__init__()
        self.enc = Standard_Transformer(dim_input, num_classes,
                                        dim_hidden=dim_hidden, ln=ln)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
        
class Tree_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1, extra_deep = False):
        super(Tree_Segmentation_Network, self).__init__()
        self.enc = Tree_Set_Transformer_Encoder(dim_input, num_classes, num_leafs,
                        ln=ln, max_tree_construction=max_tree_construction, tree=tree,
                        num_trees=num_trees, dim_hidden=dim_hidden, num_heads=num_heads,
                        extra_deep=extra_deep)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class Enhanced_Tree_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Enhanced_Tree_Segmentation_Network, self).__init__()
        self.enc = Enhanced_Tree_Set_Transformer_Encoder(dim_input, num_classes, num_leafs,
                        ln=ln, max_tree_construction=max_tree_construction, tree=tree,
                        num_trees=num_trees, dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class IV_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, dim_hidden=128,
                 num_heads=8, ln=False, num_inds=32):
        super(IV_Segmentation_Network, self).__init__()
        self.enc = Set_Transformer_Encoder(dim_input, num_classes, ln=ln,
                num_inds=num_inds, dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class IV_Segmentation_Network_deep(nn.Module):
    
    def __init__(self, dim_input, num_classes, dim_hidden=128,
                 num_heads=8, ln=False, num_inds=32):
        super(IV_Segmentation_Network_deep, self).__init__()
        self.enc = Set_Transformer_Encoder_deep(dim_input, num_classes, ln=ln,
                num_inds=num_inds, dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class Full_Segmentation_Network(nn.Module):
    
    def __init__(self, dim_input, num_classes, dim_hidden=128,
                 num_heads=8, ln=False):
        super(Full_Segmentation_Network, self).__init__()
        self.enc = Full_Set_Transformer_Encoder(dim_input, num_classes, ln=ln,
                                        dim_hidden=dim_hidden, num_heads=num_heads)
        self.softmax = nn.Softmax(dim=-1)
        
    def forward(self, X):
        X = self.enc(X)
        return self.softmax(X)
    
class Classification_Loss(nn.Module):
    
    def __init__(self):
        super(Classification_Loss, self).__init__()
        self.nll = nn.NLLLoss()
        
    def forward(self, probabilities, labels):
        probabilities = torch.log(probabilities)
        loss = self.nll(probabilities.reshape(
            [probabilities.shape[0]*probabilities.shape[1], probabilities.shape[2]]),
            labels.reshape([labels.shape[0]*labels.shape[1]]))
        return loss
    
class Test_Loss():
    
    def __init__(self):
        pass
    
    def compute_loss(self, predicted_labels, true_labels):
        loss = 0
        return loss
    
class Accuracy_Loss(Test_Loss):
    
    def __init__(self):
        super(Accuracy_Loss, self).__init__()
    
    def compute_loss(self, predicted_labels, true_labels):
        loss = torch.sum(torch.logical_not(predicted_labels == true_labels))
        loss = torch.div(loss, true_labels.shape[0])
        return loss
    
class IoU_Loss():
    
    def __init__(self):
        super(IoU_Loss, self).__init__()
    
    def compute_loss(self, predicted_labels, true_labels):
        unique_labels = torch.unique(true_labels)
        iou = torch.tensor([0], dtype=torch.float, device=device)
        
        for label in unique_labels:
            predicted_labels_mask = predicted_labels == label
            true_labels_mask = true_labels == label
            intersection = torch.logical_and(predicted_labels_mask,
                                             true_labels_mask).sum()
            union = torch.logical_or(predicted_labels_mask, true_labels_mask).sum()
            iou += torch.div(intersection, union)
            
        iou = torch.div(iou, unique_labels.shape[0])
        return iou
        

if __name__ == '__main__':
    
    D = 3
    C = 5
    l = 64
    
    B = 10
    N = 1000
    
    tree_model = Tree_Segmentation_Network(D, C, l, num_heads=4)
    iv_model = IV_Segmentation_Network(D, C, l)
    
    inp = torch.randn((B, N, D)).to(device)
    labels = torch.randint(0, C, (B, N))
    print(labels.shape)
    
    loss_fn = nn.NLLLoss()
    
    tree_classification = tree_model(inp)
    iv_classification = iv_model(inp)
    print(tree_classification)
    print(iv_classification)
    
    tree_loss = loss_fn(torch.log(tree_classification.reshape([B*N, C])), labels.reshape([B*N]))
    iv_loss = loss_fn(torch.log(iv_classification.reshape([B*N, C])), labels.reshape([B*N]))
    print(tree_loss)
    print(iv_loss)
    
    loss_fn = Classification_Loss()
    
    tree_loss = loss_fn(tree_classification, labels)
    iv_loss = loss_fn(iv_classification, labels)
    print(tree_loss)
    print(iv_loss)
    
    