#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 20:06:06 2021

@author: sandro
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math

if __name__ == '__main__':
    from tree_attention_modules_simple_tree import ISAB, SAB, PMA, Full_ISAB,\
    Tree_ISAB, reverse_ISAB, Tree_SAB
else:
    from .tree_attention_modules_simple_tree import ISAB, SAB, PMA, Full_ISAB,\
    Tree_ISAB, reverse_ISAB, Tree_SAB

class Set_Transformer(nn.Module):
    
    def __init__(self, dim_input, num_outputs, dim_output,
            num_inds=32, dim_hidden=128, num_heads=4, ln=False):
        super(Set_Transformer, self).__init__()
        self.enc = nn.Sequential(
                ISAB(dim_input, dim_hidden, num_heads, num_inds, ln=ln),
                ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln))
        self.dec = nn.Sequential(
                PMA(dim_hidden, num_heads, num_outputs, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                nn.Linear(dim_hidden, dim_output))

    def forward(self, X):
        return self.dec(self.enc(X))
    
class Full_Set_Transformer(nn.Module):
    
    def __init__(self, dim_input, num_outputs, dim_output, dim_hidden=128,
                 num_heads=4, ln=False):
        super(Full_Set_Transformer, self).__init__()
        self.enc = nn.Sequential(
                Full_ISAB(dim_input, dim_hidden, num_heads, ln=ln),
                Full_ISAB(dim_hidden, dim_hidden, num_heads, ln=ln))
        self.dec = nn.Sequential(
                PMA(dim_hidden, num_heads, num_outputs, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        return self.dec(self.enc(X))
        
class Tree_Set_Transformer(nn.Module):
    
    def __init__(self, dim_input, num_outputs, dim_output,
            num_leafs, dim_hidden=128, num_heads=4, ln=False,
            max_tree_construction=False, tree='simple', num_trees=1):
        super(Tree_Set_Transformer, self).__init__()
        self.Tree_ISAB_1 = nn.Sequential(
            Tree_ISAB(dim_input, dim_hidden, num_heads, num_leafs, ln=ln,
                      max_tree_construction=max_tree_construction, tree=tree,
                      num_trees=num_trees))
        self.Tree_ISAB_2 = nn.Sequential(
            Tree_ISAB(dim_hidden, dim_hidden, num_heads, num_leafs, ln=ln,
                      max_tree_construction=max_tree_construction, tree=tree,
                      num_trees=num_trees))
        self.dec = nn.Sequential(
                PMA(dim_hidden, num_heads, num_outputs, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        X = self.Tree_ISAB_2(self.Tree_ISAB_1(X))
        X = self.dec(X)
        return X
    
class Tree_Set_enc_dec(nn.Module):
    
    def __init__(self, dim_input, num_outputs, dim_output,
            num_leafs, dim_hidden=128, num_heads=4, ln=False,
            max_tree_construction=False, tree='simple', num_trees=1):
        super(Tree_Set_enc_dec, self).__init__()
        
        self.Encoder = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree))
        self.dec = nn.Sequential(
                PMA(dim_hidden, num_heads, num_outputs, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        X = self.Encoder(X)
        X = self.dec(X)
        return X
    
class Full_Set_enc_dec(nn.Module):
    
    def __init__(self, dim_input, num_outputs, dim_output,
            dim_hidden=128, num_heads=4, ln=False):
        super(Full_Set_enc_dec, self).__init__()
        
        self.Encoder = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads, ln=ln))
        self.dec = nn.Sequential(
                PMA(dim_hidden, num_heads, num_outputs, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                SAB(dim_hidden, dim_hidden, num_heads, ln=ln),
                nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        X = self.Encoder(X)
        X = self.dec(X)
        return X
    
class Tree_Set_Transformer_Encoder(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1, extra_deep = False):
        super(Tree_Set_Transformer_Encoder, self).__init__()
        dim_hidden_last = max(int(dim_hidden/(2*num_heads)), 4)*num_heads
        if extra_deep == False:
            self.Encoder = nn.Sequential(
                Tree_ISAB(dim_input, dim_hidden, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                Tree_ISAB(dim_hidden, dim_hidden, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                Tree_ISAB(dim_hidden, dim_hidden_last, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                nn.Linear(dim_hidden_last, dim_output))
        else:
            self.Encoder = nn.Sequential(
                Tree_ISAB(dim_input, dim_hidden, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                Tree_ISAB(dim_hidden, dim_hidden, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                Tree_ISAB(dim_hidden, dim_hidden, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                Tree_ISAB(dim_hidden, dim_hidden_last, num_heads, num_leafs, ln=ln,
                          max_tree_construction=max_tree_construction, tree=tree,
                          num_trees=num_trees),
                nn.Linear(dim_hidden_last, dim_output))
        
    def forward(self, X):
        X = self.Encoder(X)
        return X
    
class Enhanced_Tree_Set_Transformer_Encoder(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Enhanced_Tree_Set_Transformer_Encoder, self).__init__()
        dim_hidden_last = max(int(dim_hidden/(2*num_heads)), 4)*num_heads
        self.Tree_ISAB_1 = nn.Sequential(
            Tree_ISAB(dim_input, dim_hidden, num_heads, num_leafs, ln=ln,
                      max_tree_construction=max_tree_construction, tree=tree,
                      num_trees=num_trees))
        self.Tree_ISAB_2 = nn.Sequential(
            Tree_ISAB(dim_hidden+dim_input, dim_hidden, num_heads, num_leafs, ln=ln,
                      max_tree_construction=max_tree_construction, tree=tree,
                      num_trees=num_trees))
        self.Tree_ISAB_3 = nn.Sequential(
            Tree_ISAB(dim_hidden+dim_input, dim_hidden_last, num_heads, num_leafs, ln=ln,
                      max_tree_construction=max_tree_construction, tree=tree,
                      num_trees=num_trees))
        self.final = nn.Linear(dim_hidden_last, dim_output)
        
    def forward(self, X):
        Y = self.Tree_ISAB_1(X)
        Y = torch.cat((Y, X), dim=-1)
        Y = self.Tree_ISAB_2(Y)
        Y = torch.cat((Y, X), dim=-1)
        Y = self.Tree_ISAB_3(Y)
        Y = self.final(Y)
        return Y
        
class Set_Transformer_Encoder(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_inds=32, dim_hidden=128,
                 num_heads=8, ln=False):
        super(Set_Transformer_Encoder, self).__init__()
        dim_hidden_last = max(int(dim_hidden/(2*num_heads)), 4)*num_heads
        self.ISAB_1 = ISAB(dim_input, dim_hidden, num_heads, num_inds, ln=ln)
        self.ISAB_2 = ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln)
        self.ISAB_3 = ISAB(dim_hidden, dim_hidden_last, num_heads, num_inds, ln=ln)
        self.final = nn.Linear(dim_hidden_last, dim_output)

    def forward(self, X):
        X = self.ISAB_3(self.ISAB_2(self.ISAB_1(X)))
        X = self.final(X)
        return X
    
class Set_Transformer_Encoder_deep(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_inds=32, dim_hidden=128,
                 num_heads=8, ln=False):
        super(Set_Transformer_Encoder_deep, self).__init__()
        dim_hidden_last = max(int(dim_hidden/(2*num_heads)), 4)*num_heads
        self.enc = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln),
            ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln),
            ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln),
            ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln),
            ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln),
            ISAB(dim_hidden, dim_hidden_last, num_heads, num_inds, ln=ln),
            nn.Linear(dim_hidden_last, dim_output))

    def forward(self, X):
        X = self.enc(X)
        return X
    
class Set_Transformer_Encoder_plus_SLP(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_inds=32, dim_hidden=128,
                 num_heads=8, ln=False):
        super(Set_Transformer_Encoder, self).__init__()
        dim_hidden_last = max(int(dim_hidden/(2*num_heads)), 4)*num_heads
        self.ISAB_1 = ISAB(dim_input, dim_hidden, num_heads, num_inds, ln=ln)
        self.ISAB_2 = ISAB(dim_hidden, dim_hidden, num_heads, num_inds, ln=ln)
        self.ISAB_3 = reverse_ISAB(dim_hidden, dim_hidden_last, num_heads,
                                   num_inds, ln=ln)
        self.slp = torch.nn.Sequential(
                                torch.nn.Linear(dim_input, dim_hidden),
                                torch.nn.ReLU(),
                                torch.nn.Linear(dim_hidden, dim_hidden))
        self.final = nn.Linear(dim_hidden_last, dim_output)

    def forward(self, X):
        Y = self.ISAB_2(self.ISAB_1(X))
        X = self.ISAB_3(Y, self.slp(X))
        X = self.final(X)
        return X
    
class Full_Set_Transformer_Encoder(nn.Module):
    
    def __init__(self, dim_input, dim_output, dim_hidden=128,
                 num_heads=8, ln=False):
        super(Full_Set_Transformer_Encoder, self).__init__()
        dim_hidden_last = max(int(dim_hidden/(2*num_heads)), 4)*num_heads
        self.ISAB_1 = Full_ISAB(dim_input, dim_hidden, num_heads, ln=ln)
        self.ISAB_2 = Full_ISAB(dim_hidden, dim_hidden, num_heads, ln=ln)
        self.ISAB_3 = Full_ISAB(dim_hidden, dim_hidden_last, num_heads, ln=ln)
        self.final = nn.Linear(dim_hidden_last, dim_output)

    def forward(self, X):
        X = self.ISAB_2(self.ISAB_1(X))
        X = self.ISAB_3(X)
        X = self.final(X)
        return X
    
class Standard_Tree_Transformer(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1, att_softmax=True):
        super(Standard_Tree_Transformer, self).__init__()
                
        self.Encoder = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree,
                     att_softmax=att_softmax),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree,
                     att_softmax=att_softmax),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree,
                     att_softmax=att_softmax),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree,
                     att_softmax=att_softmax),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree,
                     att_softmax=att_softmax),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree,
                     att_softmax=att_softmax),
            nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        X = self.Encoder(X)
        return X
    
class Deep_Standard_Tree_Transformer(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Deep_Standard_Tree_Transformer, self).__init__()
                
        self.Encoder1 = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree))
        self.Encoder2 = nn.Sequential(
            Tree_SAB(dim_hidden+dim_input, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        Y = self.Encoder1(X)
        Y = torch.cat((Y, X), dim=-1)
        Y = self.Encoder2(Y)
        return Y
    
class Shallow_Standard_Tree_Transformer(nn.Module):
    
    def __init__(self, dim_input, dim_output, num_leafs, dim_hidden=128,
                 num_heads=8, ln=False, max_tree_construction=True, tree='simple',
                 num_trees=1):
        super(Shallow_Standard_Tree_Transformer, self).__init__()
                
        self.Encoder = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            Tree_SAB(dim_hidden, dim_hidden, num_heads=num_heads,
                     num_leafs=num_leafs, num_trees=num_trees, ln=ln,
                     max_tree_construction=max_tree_construction, tree=tree),
            nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        X = self.Encoder(X)
        return X
    
class Standard_Transformer(nn.Module):
    
    def __init__(self, dim_input, dim_output, dim_hidden=128,
                 num_heads=8, ln=False):
        super(Standard_Transformer, self).__init__()
                
        self.Encoder = nn.Sequential(
            nn.Linear(dim_input, dim_hidden),
            SAB(dim_hidden, dim_hidden, num_heads=num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads=num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads=num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads=num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads=num_heads, ln=ln),
            SAB(dim_hidden, dim_hidden, num_heads=num_heads, ln=ln),
            nn.Linear(dim_hidden, dim_output))
        
    def forward(self, X):
        X = self.Encoder(X)
        return X
