#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 13 18:39:04 2022

@author: sandro
"""

import argparse
import sys
import pickle

import numpy as np
import os
import matplotlib.pyplot as plt

if __name__ == '__main__':
    plot_dir = '../plots'
else:
    plot_dir = 'plots'

class Custom_Color_Cycle():
    
    def __init__(self, colors):
        
        self.colors = colors
        self.length = len(colors)
        
        self.counter = 0
        
    def get_color(self):
        
        color = self.colors[self.counter]
        self.counter += 1
        self.counter = self.counter % self.length
        
        return color
    
    def reset(self):
        
        self.counter = 0
        
custom_color_cycle = Custom_Color_Cycle(['b', 'r', 'g', 'c', 'm', 'y'])

def mean_average_plot(data_dict, plot_file, title, x_label, y_label, x_limits=None,
                      y_limits=None):
    
    if x_limits == None:
        x_min = data_dict[list(data_dict)[0]][0][0]
        x_max = x_min
        for key in data_dict:
            epchs = data_dict[key][0]
            x_min = min(np.min(epchs), x_min)
            x_max = max(np.max(epchs), x_max)
        x_limits = [x_min, x_max]
    
    fig, ax = plt.subplots(figsize=(14,10))
    ax.set_title(title, fontsize=30)
    ax.set_xlim(x_limits)
    if y_limits != None:
        ax.set_ylim(y_limits)
    ax.set_ylabel(y_label, fontsize=20)
    ax.set_xlabel(x_label, fontsize=20)
    ax.tick_params(axis='both', labelsize=15)
    
    # plot all triples in data_dict
    custom_color_cycle.reset()
    for key in data_dict:
        color = custom_color_cycle.get_color()
        epchs = data_dict[key][0]
        mean_performances = data_dict[key][1]
        std_devs = data_dict[key][2]
        lineplot = plt.plot(epchs, mean_performances, color=color,
                            label=key)
        ax.fill_between(epchs, np.subtract(mean_performances, std_devs),
                        np.add(mean_performances, std_devs),
                        color=color, alpha=0.3)
        
    ax.legend(fontsize=20)
    path = os.path.join(plot_dir, plot_file)
    plt.savefig(path)
    plt.show()

    
if __name__ == '__main__':
    
    # parse command line arguments
    parser = argparse.ArgumentParser(description='Specify input files, axis descriptions'+\
                                     ', etc. for plotting')
    parser.add_argument('-run_performances', metavar='run_performances', nargs='+',
                        type=str, default=None, help='specify the saved run ' +\
                            'performances to plot')
    parser.add_argument('-models', metavar='models', nargs='+',
                        type=str, default=None, help='specify whether to plot ' +\
                            'the model or the benchmark performance or both')
    parser.add_argument('-curve_labels', metavar='curve_labels', nargs='+',
                        type=str, default=None, help='specify the labels for ' +\
                            'the plotted curves')
    parser.add_argument('-plot_title', metavar='plot_title', type=str, default='plot',
                        help='specify the title of the plot')
    parser.add_argument('-x_label', metavar='x_label', type=str, default='x label',
                        help='specify the label of the x axis')
    parser.add_argument('-y_label', metavar='y_label', type=str, default='y label',
                        help='specify the label of the y axis')
    parser.add_argument('-x_min', metavar='x_min', type=int, default=None,
                        help='specify the min limit of the x axis')
    parser.add_argument('-x_max', metavar='x_max', type=int, default=None,
                        help='specify the max limit of the x axis')
    parser.add_argument('-y_min', metavar='y_min', type=int, default=None,
                        help='specify the min limit of the y axis')
    parser.add_argument('-y_max', metavar='y_max', type=int, default=None,
                        help='specify the max limit of the y axis')
    parser.add_argument('-plot_filename', metavar='plot_filename', type=str,
                        default='unnamed_plot',
                        help='specify the name of the saved plot png-file')
    
    
    args = parser.parse_args()
    run_performance_files = args.run_performances
    models = args.models
    curve_labels = args.curve_labels
    title = args.plot_title
    x_label = args.x_label
    y_label = args.y_label
    x_min = args.x_min
    x_max = args.x_max
    y_min = args.y_min
    y_max = args.y_max
    plot_file = args.plot_filename + '.png'
    
    if x_min != None and x_max != None:
        x_min = min(x_min, x_max)
        x_limits = [x_min, x_max]
    else:
        x_limits = None
        
    if y_min != None and y_max != None:
        y_min = min(y_min, y_max)
        y_limits = [y_min, y_max]
    else:
        y_limits = None
    
    if models == None:
        models = []
        
    if curve_labels == None:
        curve_labels = []
        
    if run_performance_files == None:
        print('Please specify at least one datafile to load!')
        sys.exit()
        
    plot_modes = ['both']*len(run_performance_files)
    for i in range(len(models)):
        plot_modes[i] = models[i]
        
    data_dict = {}
    file_idx = 0
    label_idx = 0
    label_escape_number = 0
    for datafile in run_performance_files:
        
        datafile = '../run_performances/' + datafile
        perf_file = open(datafile, "rb")
        perf_dict = pickle.load(perf_file)
        perf_file.close()
        
        if plot_modes[file_idx] in ['model', 'both']:
            epochs = perf_dict['model epochs']
            mean = perf_dict['model mean performances']
            std_dev = perf_dict['model performances std devs']
            if label_idx < len(curve_labels):
                label = curve_labels[label_idx]
                label_idx += 1
            else:
                label = perf_dict['model label']
            if label in data_dict:
                label = label + str(label_escape_number)
                label_escape_number += 1
            plot_list = [epochs, mean, std_dev]
            data_dict.update({label: plot_list})
            
        if plot_modes[file_idx] in ['benchmark', 'both']:
            epochs = perf_dict['benchmark epochs']
            mean = perf_dict['benchmark mean performances']
            std_dev = perf_dict['benchmark performances std devs']
            if label_idx < len(curve_labels):
                label = curve_labels[label_idx]
                label_idx += 1
            else:
                label = perf_dict['benchmark label']
            if label in data_dict:
                label = label + str(label_escape_number)
                label_escape_number += 1
            plot_list = [epochs, mean, std_dev]
            data_dict.update({label: plot_list})
        
        file_idx += 1
    
    mean_average_plot(data_dict, plot_file, title, x_label, y_label, x_limits,
                      y_limits)
    
    
    