#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  4 20:24:40 2022

@author: sandro
"""

import torch
import math
import argparse
import sys
import pickle

import numpy as np
import os
from time import time, sleep
from torch.utils import tensorboard
from omegaconf import OmegaConf
from datetime import datetime

from Particle_Movement import Particle_Movement_Tools

import sys 
sys.path.append('..')
from Tree_attention_scaling_simple_tree_no_sparse.tree_attention_models\
    import Standard_Tree_Transformer, Deep_Standard_Tree_Transformer,\
        Shallow_Standard_Tree_Transformer, Standard_Transformer

device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

run_performances_folder = 'run_performances'
# be aware of the running time in order to avoid job cancellations due to time limit
job_launch_time = time()
time_margin = 600

# parse command line arguments
parser = argparse.ArgumentParser(description='Parameters for point cloud '+\
                                 'segmentation pipeline')
parser.add_argument('-batch_size', metavar='batch_size', type=int,
                    default=16, help='Training Batch Size')
parser.add_argument('-n_particles', metavar='n_particles', type=int,
                    default=100, help='Number of particles')
parser.add_argument('-simulation_type', metavar='simulation_type', type=str,
                    default='gravity', help='Force simulation type (gravity or coulomb)')
parser.add_argument('-model', metavar='model', type=str,
                    default='standard', help='Model')
parser.add_argument('-tree_depth', metavar='tree_depth', type=int,
                    default=5, help='Depth of the attention tree')
parser.add_argument('-ln', metavar='ln', type=bool,
                    default=True, help='Whether to do layer normalization or not')
parser.add_argument('-lr', metavar='lr', type=float,
                    default=1e-3, help='Learning rate')
parser.add_argument('-lr_decay', metavar='lr_decay', type=float,
                    default=0.97725, help='Per epoch learning rate decay factor')
parser.add_argument('-training_epochs', metavar='training_epochs', type=int,
                    default=200, help='Number of training epochs')
parser.add_argument('-smooth_lr_decay', metavar='smooth_lr_decay', type=bool,
                    default=False, help='whether to let the lr decay smoothly or not')
parser.add_argument('-att_softmax', metavar='att_softmax', type=bool,
                    default=False, help='whether to use softmax in attention calculation or not')

args = parser.parse_args()
batch_size = args.batch_size
n_particles = args.n_particles
simulation_type = args.simulation_type
model_architecture = args.model
tree_depth = args.tree_depth
ln = args.ln
lr = args.lr
lr_decay = args.lr_decay
training_epochs = args.training_epochs
smooth_lr_decay = args.smooth_lr_decay
att_softmax = args.att_softmax

if not n_particles in [3, 10, 20, 50, 100, 1000]:
    print('ERROR: number of particles not in [3, 10, 20, 50, 100, 1000]')
    
if not simulation_type in ['gravity', 'coulomb']:
    print('ERROR: simulation_type not in [\'gravity\', \'coulomb\']')
elif simulation_type == 'gravity':
    particle_dimension = 5
elif simulation_type == 'coulomb':
    particle_dimension = 6
    
if not model_architecture in ['shallow', 'standard', 'deep', 'dense_standard']:
    print('ERROR: model_depth not in [\'shallow\', \'standard\', \'deep\',' +\
          '\'dense_standard\']')
    
# dirty solution to circumvent occasionally appearing torch script error
def create_model(particle_dimension, num_leafs, ln, max_tree_construction=False,
                 tree='plusplus', num_trees=1, num_heads=8, dim_hidden=128):
    try:
        if model_architecture == 'standard':
            model = Standard_Tree_Transformer(particle_dimension,
                    particle_dimension-1, num_leafs, ln=ln,
                    max_tree_construction=max_tree_construction,
                    tree=tree, num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden, att_softmax=att_softmax).to(device)
        elif model_architecture == 'deep':
            model = Deep_Standard_Tree_Transformer(particle_dimension,
                    particle_dimension-1, num_leafs, ln=ln,
                    max_tree_construction=max_tree_construction,
                    tree=tree, num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
        elif model_architecture == 'shallow':
            model = Shallow_Standard_Tree_Transformer(particle_dimension,
                    particle_dimension-1, num_leafs, ln=ln,
                    max_tree_construction=max_tree_construction,
                    tree=tree, num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
        elif model_architecture == 'dense_standard':
            model = Standard_Transformer(particle_dimension,
                    particle_dimension-1, ln=ln, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
    except:
        sleep(1)
        print('retrying to create model')
        model = create_model(particle_dimension, num_leafs, ln, max_tree_construction,
                             tree, num_trees, num_heads, dim_hidden)
    return model

if __name__ == '__main__':
    
    dataset_file = str(n_particles) + '_particles_' + simulation_type
    
    model = create_model(particle_dimension, 2**tree_depth, ln)
    
    Particle_Movement_Tools.train_model(model, dataset=dataset_file, learning_rate=lr,
                                        lr_decay=lr_decay, batch_size=batch_size,
                                        data_dir="Particle_Movement/data",
                                        epochs=training_epochs, smooth_lr_decay=smooth_lr_decay)


